@@ -1,31 +1,33 @@
 """Tests for parallel client.py"""
-
 # Copyright (c) IPython Development Team.
 # Distributed under the terms of the Modified BSD License.
-
 from __future__ import division
 
+import os
+import signal
 import socket
+import sys
+import time
 from concurrent.futures import Future
 from datetime import datetime
-import os
-import sys
 from threading import Thread
-import time
-try:
-    from unittest import mock
-except ImportError:
-    import mock
+from unittest import mock
 
 import pytest
 import tornado
-
 from IPython import get_ipython
+
+from .clienttest import ClusterTestCase
+from .clienttest import raises_remote
+from .clienttest import skip_without
+from .clienttest import wait
+from ipyparallel import AsyncHubResult
+from ipyparallel import DirectView
+from ipyparallel import error
+from ipyparallel import Reference
 from ipyparallel.client import client as clientmod
-from ipyparallel import error, AsyncHubResult, DirectView, Reference
 from ipyparallel.util import utc
 
-from .clienttest import ClusterTestCase, wait, add_engines, skip_without
 
 @pytest.mark.usefixtures('ipython')
 class TestClient(ClusterTestCase):
@@ -35,12 +37,11 @@ class TestClient(ClusterTestCase):
     def test_ids(self):
         n = len(self.client.ids)
         self.add_engines(2)
-        self.assertEqual(len(self.client.ids), n+2)
-
+        self.assertEqual(len(self.client.ids), n + 2)
 
     def test_iter(self):
         self.minimum_engines(4)
-        engine_ids = [ view.targets for view in self.client ]
+        engine_ids = [view.targets for view in self.client]
         self.assertEqual(engine_ids, self.client.ids)
 
     def test_view_indexing(self):
@@ -69,7 +70,20 @@ class TestClient(ClusterTestCase):
         v = self.client[-1]
         self.assertTrue(isinstance(v, DirectView))
         self.assertEqual(v.targets, targets[-1])
-        self.assertRaises(TypeError, lambda : self.client[None])
+        self.assertRaises(TypeError, lambda: self.client[None])
+
+    def test_outstanding(self):
+        self.minimum_engines(1)
+        e = self.client[-1]
+        engine_id = self.client._engines[e.targets]
+        ar = e.apply_async(time.sleep, 0.5)
+        msg_id = ar.msg_ids[0]
+        # verify that msg id
+        assert msg_id in self.client.outstanding
+        assert msg_id in self.client._outstanding_dict[engine_id]
+        ar.get()
+        assert msg_id not in self.client.outstanding
+        assert msg_id not in self.client._outstanding_dict[engine_id]
 
     def test_lbview_targets(self):
         """test load_balanced_view targets"""
@@ -95,16 +109,17 @@ class TestClient(ClusterTestCase):
         self.assertEqual(v.targets, 'all')
 
         def double(x):
-            return x*2
+            return x * 2
+
         seq = list(range(100))
-        ref = [ double(x) for x in seq ]
+        ref = [double(x) for x in seq]
 
         # add some engines, which should be used
         self.add_engines(1)
         n1 = len(self.client.ids)
 
         # simple apply
-        r = v.apply_sync(lambda : 1)
+        r = v.apply_sync(lambda: 1)
         self.assertEqual(r, [1] * n1)
 
         # map goes through remotefunction
@@ -117,7 +132,7 @@ class TestClient(ClusterTestCase):
         self.assertNotEqual(n2, n1)
 
         # apply
-        r = v.apply_sync(lambda : 1)
+        r = v.apply_sync(lambda: 1)
         self.assertEqual(r, [1] * n2)
 
         # map
@@ -128,14 +143,14 @@ class TestClient(ClusterTestCase):
         """test various valid targets arguments"""
         build = self.client._build_targets
         ids = self.client.ids
-        idents,targets = build(None)
+        idents, targets = build(None)
         self.assertEqual(ids, targets)
 
     def test_clear(self):
         """test clear behavior"""
         self.minimum_engines(2)
         v = self.client[:]
-        v.block=True
+        v.block = True
         v.push(dict(a=5))
         v.pull('a')
         id0 = self.client.ids[-1]
@@ -152,7 +167,7 @@ class TestClient(ClusterTestCase):
         t = c.ids[-1]
         ar = c[t].apply_async(wait, 1)
         # give the monitor time to notice the message
-        time.sleep(.25)
+        time.sleep(0.25)
         ahr = self.client.get_result(ar.msg_ids[0], owner=False)
         self.assertIsInstance(ahr, AsyncHubResult)
         self.assertEqual(ahr.get(), ar.get())
@@ -169,14 +184,10 @@ class TestClient(ClusterTestCase):
         """test getting execute results from the Hub."""
         c = clientmod.Client(profile='iptest')
         t = c.ids[-1]
-        cell = '\n'.join([
-            'import time',
-            'time.sleep(0.25)',
-            '5'
-        ])
+        cell = '\n'.join(['import time', 'time.sleep(0.25)', '5'])
         ar = c[t].execute("import time; time.sleep(1)", silent=False)
         # give the monitor time to notice the message
-        time.sleep(.25)
+        time.sleep(0.25)
         ahr = self.client.get_result(ar.msg_ids[0], owner=False)
         self.assertIsInstance(ahr, AsyncHubResult)
         self.assertEqual(ahr.get().execute_result, ar.get().execute_result)
@@ -210,7 +221,7 @@ class TestClient(ClusterTestCase):
         ids = sorted(self.client.ids)
         self.assertEqual(intkeys, ids)
         unassigned = allqs.pop('unassigned')
-        for eid,qs in allqs.items():
+        for eid, qs in allqs.items():
             self.assertTrue(isinstance(qs, dict))
             self.assertEqual(sorted(qs.keys()), ['completed', 'queue', 'tasks'])
 
@@ -227,7 +238,7 @@ class TestClient(ClusterTestCase):
                 break
             time.sleep(0.1)
         self.assertNotIn(id0, self.client.ids)
-        self.assertRaises(IndexError, lambda : self.client[id0])
+        self.assertRaises(IndexError, lambda: self.client[id0])
 
     def test_result_status(self):
         pass
@@ -236,49 +247,55 @@ class TestClient(ClusterTestCase):
     def test_db_query_dt(self):
         """test db query by date"""
         hist = self.client.hub_history()
-        middle = self.client.db_query({'msg_id' : hist[len(hist)//2]})[0]
+        middle = self.client.db_query({'msg_id': hist[len(hist) // 2]})[0]
         tic = middle['submitted']
-        before = self.client.db_query({'submitted' : {'$lt' : tic}})
-        after = self.client.db_query({'submitted' : {'$gte' : tic}})
-        self.assertEqual(len(before)+len(after),len(hist))
+        before = self.client.db_query({'submitted': {'$lt': tic}})
+        after = self.client.db_query({'submitted': {'$gte': tic}})
+        self.assertEqual(len(before) + len(after), len(hist))
         for b in before:
             self.assertTrue(b['submitted'] < tic)
         for a in after:
             self.assertTrue(a['submitted'] >= tic)
-        same = self.client.db_query({'submitted' : tic})
+        same = self.client.db_query({'submitted': tic})
         for s in same:
             self.assertTrue(s['submitted'] == tic)
 
     def test_db_query_keys(self):
         """test extracting subset of record keys"""
-        found = self.client.db_query({'msg_id': {'$ne' : ''}},keys=['submitted', 'completed'])
+        found = self.client.db_query(
+            {'msg_id': {'$ne': ''}}, keys=['submitted', 'completed']
+        )
         for rec in found:
             self.assertEqual(set(rec.keys()), set(['msg_id', 'submitted', 'completed']))
 
     def test_db_query_default_keys(self):
         """default db_query excludes buffers"""
-        found = self.client.db_query({'msg_id': {'$ne' : ''}})
+        found = self.client.db_query({'msg_id': {'$ne': ''}})
         for rec in found:
             keys = set(rec.keys())
             self.assertFalse('buffers' in keys, "'buffers' should not be in: %s" % keys)
-            self.assertFalse('result_buffers' in keys, "'result_buffers' should not be in: %s" % keys)
+            self.assertFalse(
+                'result_buffers' in keys, "'result_buffers' should not be in: %s" % keys
+            )
 
     def test_db_query_msg_id(self):
         """ensure msg_id is always in db queries"""
-        found = self.client.db_query({'msg_id': {'$ne' : ''}},keys=['submitted', 'completed'])
+        found = self.client.db_query(
+            {'msg_id': {'$ne': ''}}, keys=['submitted', 'completed']
+        )
         for rec in found:
             self.assertTrue('msg_id' in rec.keys())
-        found = self.client.db_query({'msg_id': {'$ne' : ''}},keys=['submitted'])
+        found = self.client.db_query({'msg_id': {'$ne': ''}}, keys=['submitted'])
         for rec in found:
             self.assertTrue('msg_id' in rec.keys())
-        found = self.client.db_query({'msg_id': {'$ne' : ''}},keys=['msg_id'])
+        found = self.client.db_query({'msg_id': {'$ne': ''}}, keys=['msg_id'])
         for rec in found:
             self.assertTrue('msg_id' in rec.keys())
 
     def test_db_query_get_result(self):
         """pop in db_query shouldn't pop from result itself"""
-        self.client[:].apply_sync(lambda : 1)
-        found = self.client.db_query({'msg_id': {'$ne' : ''}})
+        self.client[:].apply_sync(lambda: 1)
+        found = self.client.db_query({'msg_id': {'$ne': ''}})
         rc2 = clientmod.Client(profile='iptest')
         # If this bug is not fixed, this call will hang:
         ar = rc2.get_result(self.client.history[-1])
@@ -292,30 +309,30 @@ class TestClient(ClusterTestCase):
         hist = self.client.hub_history()
         even = hist[::2]
         odd = hist[1::2]
-        recs = self.client.db_query({ 'msg_id' : {'$in' : even}})
-        found = [ r['msg_id'] for r in recs ]
+        recs = self.client.db_query({'msg_id': {'$in': even}})
+        found = [r['msg_id'] for r in recs]
         self.assertEqual(set(even), set(found))
-        recs = self.client.db_query({ 'msg_id' : {'$nin' : even}})
-        found = [ r['msg_id'] for r in recs ]
+        recs = self.client.db_query({'msg_id': {'$nin': even}})
+        found = [r['msg_id'] for r in recs]
         self.assertEqual(set(odd), set(found))
 
     def test_hub_history(self):
         hist = self.client.hub_history()
-        recs = self.client.db_query({ 'msg_id' : {"$ne":''}})
+        recs = self.client.db_query({'msg_id': {"$ne": ''}})
         recdict = {}
         for rec in recs:
             recdict[rec['msg_id']] = rec
 
-        latest = datetime(1984,1,1).replace(tzinfo=utc)
+        latest = datetime(1984, 1, 1).replace(tzinfo=utc)
         for msg_id in hist:
             rec = recdict[msg_id]
             newt = rec['submitted']
             self.assertTrue(newt >= latest)
             latest = newt
-        ar = self.client[-1].apply_async(lambda : 1)
+        ar = self.client[-1].apply_async(lambda: 1)
         ar.get()
         time.sleep(0.25)
-        self.assertEqual(self.client.hub_history()[-1:],ar.msg_ids)
+        self.assertEqual(self.client.hub_history()[-1:], ar.msg_ids)
 
     def _wait_for_idle(self):
         """wait for the cluster to become idle, according to the everyone."""
@@ -342,7 +359,9 @@ class TestClient(ClusterTestCase):
         # timeout 5s, polling every 100ms
         qs = rc.queue_status()
         for i in range(50):
-            if qs['unassigned'] or any(qs[eid]['tasks'] + qs[eid]['queue'] for eid in qs if eid != 'unassigned'):
+            if qs['unassigned'] or any(
+                qs[eid]['tasks'] + qs[eid]['queue'] for eid in qs if eid != 'unassigned'
+            ):
                 time.sleep(0.1)
                 qs = rc.queue_status()
             else:
@@ -350,15 +369,16 @@ class TestClient(ClusterTestCase):
 
         # ensure Hub up to date:
         self.assertEqual(qs['unassigned'], 0)
-        for eid in [ eid for eid in qs if eid != 'unassigned' ]:
+        for eid in [eid for eid in qs if eid != 'unassigned']:
             self.assertEqual(qs[eid]['tasks'], 0)
             self.assertEqual(qs[eid]['queue'], 0)
 
-
     def test_resubmit(self):
         def f():
             import random
+
             return random.random()
+
         v = self.client.load_balanced_view()
         ar = v.apply_async(f)
         r1 = ar.get(1)
@@ -371,7 +391,7 @@ class TestClient(ClusterTestCase):
     def test_resubmit_chain(self):
         """resubmit resubmitted tasks"""
         v = self.client.load_balanced_view()
-        ar = v.apply_async(lambda x: x, 'x'*1024)
+        ar = v.apply_async(lambda x: x, 'x' * 1024)
         ar.get()
         self._wait_for_idle()
         ars = [ar]
@@ -380,13 +400,16 @@ class TestClient(ClusterTestCase):
             ar = ars[-1]
             ar2 = self.client.resubmit(ar.msg_ids)
 
-        [ ar.get() for ar in ars ]
+        [ar.get() for ar in ars]
 
     def test_resubmit_header(self):
         """resubmit shouldn't clobber the whole header"""
+
         def f():
             import random
+
             return random.random()
+
         v = self.client.load_balanced_view()
         v.retries = 1
         ar = v.apply_async(f)
@@ -396,8 +419,10 @@ class TestClient(ClusterTestCase):
         ahr = self.client.resubmit(ar.msg_ids)
         ahr.get(1)
         time.sleep(0.5)
-        records = self.client.db_query({'msg_id': {'$in': ar.msg_ids + ahr.msg_ids}}, keys='header')
-        h1,h2 = [ r['header'] for r in records ]
+        records = self.client.db_query(
+            {'msg_id': {'$in': ar.msg_ids + ahr.msg_ids}}, keys='header'
+        )
+        h1, h2 = [r['header'] for r in records]
         for key in set(h1.keys()).union(set(h2.keys())):
             if key in ('msg_id', 'date'):
                 self.assertNotEqual(h1[key], h2[key])
@@ -407,7 +432,9 @@ class TestClient(ClusterTestCase):
     def test_resubmit_aborted(self):
         def f():
             import random
+
             return random.random()
+
         v = self.client.load_balanced_view()
         # restrict to one engine, so we can put a sleep
         # ahead of the task, so it will get aborted
@@ -425,7 +452,7 @@ class TestClient(ClusterTestCase):
     def test_resubmit_inflight(self):
         """resubmit of inflight task"""
         v = self.client.load_balanced_view()
-        ar = v.apply_async(time.sleep,1)
+        ar = v.apply_async(time.sleep, 1)
         # give the message a chance to arrive
         time.sleep(0.2)
         ahr = self.client.resubmit(ar.msg_ids)
@@ -439,7 +466,7 @@ class TestClient(ClusterTestCase):
     def test_purge_hub_results(self):
         # ensure there are some tasks
         for i in range(5):
-            self.client[:].apply_sync(lambda : 1)
+            self.client[:].apply_sync(lambda: 1)
         # Wait for the Hub to realise the result is done:
         # This prevents a race condition, where we
         # might purge a result the Hub still thinks is pending.
@@ -450,7 +477,7 @@ class TestClient(ClusterTestCase):
         ahr.wait(10)
         self.client.purge_hub_results(hist[-1])
         newhist = self.client.hub_history()
-        self.assertEqual(len(newhist)+1,len(hist))
+        self.assertEqual(len(newhist) + 1, len(hist))
         rc2.close()
 
     def test_purge_local_results(self):
@@ -462,8 +489,12 @@ class TestClient(ClusterTestCase):
         before = len(self.client.results)
         self.assertEqual(len(self.client.metadata), before)
         self.client.purge_local_results(msg_id)
-        self.assertLessEqual(len(self.client.results), before-1, msg="Not removed from results")
-        self.assertLessEqual(len(self.client.metadata), before-1, msg="Not removed from metadata")
+        self.assertLessEqual(
+            len(self.client.results), before - 1, msg="Not removed from results"
+        )
+        self.assertLessEqual(
+            len(self.client.metadata), before - 1, msg="Not removed from metadata"
+        )
 
     def test_purge_local_results_outstanding(self):
         v = self.client[-1]
@@ -494,7 +525,7 @@ class TestClient(ClusterTestCase):
     def test_purge_all_results(self):
         # ensure there are some tasks
         for i in range(5):
-            self.client[:].apply_sync(lambda : 1)
+            self.client[:].apply_sync(lambda: 1)
         assert self.client.wait(timeout=10)
         self._wait_for_idle()
         self.client.purge_results('all')
@@ -506,7 +537,7 @@ class TestClient(ClusterTestCase):
     def test_purge_everything(self):
         # ensure there are some tasks
         for i in range(5):
-            self.client[:].apply_sync(lambda : 1)
+            self.client[:].apply_sync(lambda: 1)
         self.client.wait(timeout=10)
         self._wait_for_idle()
         self.client.purge_everything()
@@ -514,7 +545,9 @@ class TestClient(ClusterTestCase):
         self.assertEqual(len(self.client.results), 0, msg="Results not empty")
         self.assertEqual(len(self.client.metadata), 0, msg="metadata not empty")
         # The client "bookkeeping"
-        self.assertEqual(len(self.client.session.digest_history), 0, msg="session digest not empty")
+        self.assertEqual(
+            len(self.client.session.digest_history), 0, msg="session digest not empty"
+        )
         self.assertEqual(len(self.client.history), 0, msg="client history not empty")
         # the hub results
         hist = self.client.hub_history()
@@ -541,15 +574,17 @@ class TestClient(ClusterTestCase):
         self.assertEqual(v0.targets, 'all')
 
     def test_wait_interactive(self):
-        ar = self.client[-1].apply_async(lambda : 1)
+        ar = self.client[-1].apply_async(lambda: 1)
         self.client.wait_interactive()
         self.assertEqual(self.client.outstanding, set())
 
     def test_await_future(self):
         f = Future()
+
         def finish_later():
             time.sleep(0.1)
             f.set_result('future')
+
         Thread(target=finish_later).start()
         assert self.client.wait([f])
         assert f.done()
@@ -557,11 +592,12 @@ class TestClient(ClusterTestCase):
 
     @skip_without('distributed')
     @pytest.mark.skipif(
-        sys.version_info[:2] <= (3, 5),
-        reason="become_dask doesn't work on Python 3.5")
+        sys.version_info[:2] <= (3, 5), reason="become_dask doesn't work on Python 3.5"
+    )
     @pytest.mark.skipif(
         tornado.version_info[:2] < (5,),
-        reason="become_dask doesn't work with tornado 4")
+        reason="become_dask doesn't work with tornado 4",
+    )
     def test_become_dask(self):
         executor = self.client.become_dask()
         reprs = self.client[:].apply_sync(repr, Reference('distributed_worker'))
@@ -580,21 +616,68 @@ class TestClient(ClusterTestCase):
 
     def test_warning_on_hostname_match(self):
         location = socket.gethostname()
-        with mock.patch('ipyparallel.client.client.is_local_ip',
-                        lambda x: False):
-            with mock.patch('socket.gethostname', lambda: location[0:-1]), \
-                 pytest.warns(RuntimeWarning): # should trigger warning
-                c = self.connect_client()
+        with mock.patch('ipyparallel.client.client.is_local_ip', lambda x: False):
+            with mock.patch('socket.gethostname', lambda: location[0:-1]):
+                with pytest.warns(RuntimeWarning):  # should trigger warning
+                    c = self.connect_client()
                 c.close()
-            with mock.patch('socket.gethostname', lambda: location), \
-                    pytest.warns(None) as record:  # should not trigger warning
-                c = self.connect_client()
-                assert len(record) == 0, str([str(w) for w in record])
+            with mock.patch('socket.gethostname', lambda: location):
+                with pytest.warns(None) as record:  # should not trigger warning
+                    c = self.connect_client()
+                # only capture runtime warnings
+                runtime_warnings = [
+                    w for w in record if isinstance(w.message, RuntimeWarning)
+                ]
+                assert len(runtime_warnings) == 0, str(
+                    [str(w) for w in runtime_warnings]
+                )
                 c.close()
 
     def test_local_ip_true_doesnt_trigger_warning(self):
-        with mock.patch('ipyparallel.client.client.is_local_ip',
-                        lambda x: True), pytest.warns(None) as record:
-            c = self.connect_client()
-            assert len(record) == 0, str([str(w) for w in record])
+        with mock.patch('ipyparallel.client.client.is_local_ip', lambda x: True):
+            with pytest.warns(None) as record:
+                c = self.connect_client()
+            # only capture runtime warnings
+            runtime_warnings = [
+                w for w in record if isinstance(w.message, RuntimeWarning)
+            ]
+            assert len(runtime_warnings) == 0, str([str(w) for w in runtime_warnings])
             c.close()
+
+    def test_wait_for_engines(self):
+        n = len(self.client)
+        assert self.client.wait_for_engines(n) is None
+
+        f = self.client.wait_for_engines(n, block=False)
+        assert isinstance(f, Future)
+        assert f.done()
+
+        with pytest.raises(TimeoutError):
+            self.client.wait_for_engines(n + 1, timeout=0.1)
+
+        f = self.client.wait_for_engines(n + 1, timeout=10, block=False)
+        self.add_engines(1)
+        assert f.result() is None
+
+    @pytest.mark.skipif(
+        sys.platform.startswith("win"), reason="Signal tests don't pass on Windows yet"
+    )
+    def test_signal_engines(self):
+
+        view = self.client[:]
+        if sys.platform.startswith("win"):
+            signame = 'CTRL_C_EVENT'
+        else:
+            signame = 'SIGINT'
+        signum = getattr(signal, signame)
+        for sig in (signum, signame):
+            ar = view.apply_async(time.sleep, 10)
+            # FIXME: use status:busy to wait for tasks to start
+            time.sleep(1)
+            self.client.send_signal(sig, block=True)
+            with raises_remote(KeyboardInterrupt):
+                ar.get()
+
+            # make sure they were all interrupted
+            for r in ar.get(return_exceptions=True):
+                assert isinstance(r, error.RemoteError)
